#==============================================================================

# Copyright (c) 2016-10-21, Laurenz Wiskott, http://www.ini.rub.de/PEOPLE/wiskott/
# License: BSD 2-Clause, https://opensource.org/licenses/BSD-2-Clause, see the end

# Inspired by
# http://scikit-learn.org/stable/auto_examples/decomposition/plot_faces_decomposition.html

#==============================================================================

import numpy as np
import matplotlib.pyplot as plt
import time

from sklearn.datasets import fetch_olivetti_faces
from sklearn import decomposition

IMAGE_SHAPE = (64, 64)
SHOW_BLOCKS = True # Program should stop at each figure or run through.
CENTER_FACES = True # Faces should be centered or not.

#==============================================================================

# PLOTTING ROUTINES

def show_gallery(title, images, num_col=1, num_row=1):
    """
    Routine to plot a set of square images.
    """
 
    unit = 1.8 # Change this unit if you want the figures larger or smaller.

    top_space    = 0.40*unit
    left_space   = 0.10*unit
    bottom_space = 0.10*unit
    right_space  = left_space
    image_height = 1.00*unit
    image_hspace = 0.05*unit
    image_width  = image_height
    image_wspace = image_hspace

    fig_height = num_row*image_height + (num_row-1)*image_hspace +  top_space + bottom_space
    fig_width  = num_col*image_width  + (num_col-1)*image_wspace + left_space +  right_space

    fig = plt.figure(figsize=(fig_width, fig_height))
    fig.suptitle(title, size=int(15*unit))
    for i, image in enumerate(images):
        ax = fig.add_subplot(num_row, num_col, i + 1)
        ax.imshow(image.reshape(IMAGE_SHAPE), cmap=plt.cm.gray,
                   interpolation='nearest'
        )
        ax.xaxis.set_ticks([])
        ax.yaxis.set_ticks([])
    fig.subplots_adjust(left=left_space/fig_width, bottom=bottom_space/fig_height,
                        right=1-right_space/fig_width, top=1-top_space/fig_height,
                        wspace=(num_row-1)*image_wspace/fig_width,
                        hspace=(num_col-1)*image_hspace/fig_height)
    plt.show(block=SHOW_BLOCKS)


def animate_gallery(images, stride=1):
    """
    Routine to animate a set of images.
    """

    fig = plt.figure(figsize=(2.,2.26))
    fig.suptitle("Animation of faces", size=16)
    plt.show(block=False)
    ax = fig.add_subplot(111)
    for i, image in enumerate(images):
        if not i%stride == 0:
            continue
        ax.imshow(image.reshape(IMAGE_SHAPE), cmap=plt.cm.gray,
                  interpolation='nearest')
        ax.xaxis.set_ticks([])
        ax.yaxis.set_ticks([])
        fig.canvas.draw()    
    plt.show(block=SHOW_BLOCKS)


def animate_proj(faces_proj, comp_1=0, comp_2=1, sleep=0.2):
    """
    Routine to animate projected points of one id within all points of all ids.
    """

    fig = plt.figure(figsize=(10,10))
    fig.suptitle("Animation of projected face images in PCA components "+str(comp_1)+" and "+str(comp_2), size=16)
    plt.show(block=False)
    ax = fig.add_subplot(111)
    for id_idx in range(40):
        ax.clear()
        ax.scatter(faces_proj[:,comp_1],
                   faces_proj[:,comp_2],
                   s=10, marker='.')
        ax.scatter(faces_proj[id_idx*10:(id_idx+1)*10,comp_1],
                   faces_proj[id_idx*10:(id_idx+1)*10,comp_2], 
                   s=50, color='r', marker='o')
        fig.canvas.draw()
        time.sleep(sleep)
    plt.show(block=SHOW_BLOCKS)


#==============================================================================

# LOAD FACE IMAGES

# The Olivetti face dataset used here consists of 10 slightly different views
# or expressions of 40 faces, each image with 64 x 64 pixels.  The images are
# aligned at the pupils of the eyes.  They are stored in a 400 x 4096 array.

faces = fetch_olivetti_faces().data
print "faces.shape =", faces.shape

# Plot all 10 views of the first five faces in the data base.
print "Show all raw images of first five individuals in an overview plot."
show_gallery("All views of first five faces", faces[:50], num_col=10, num_row=5)

# Animate gallery of face to see how well aligned the faces are. 
print "Show all raw images of first five individuals in an animation plot to see the alignment at the eyes."
animate_gallery(faces, 5)

#==============================================================================

# EXERCISES

# Explore the data with the help of PCA.  Start without looking at the
# exercises below.  Try to be creative here.  If you run out of ideas, look
# below.

#------------------------------------------------------------------------------

# EXERCISE 1: Center the face images by subtracting the mean face image.

# SOLUTION:
if CENTER_FACES:
    faces_mean = faces.mean(axis=0)
    print "faces_mean.shape =", faces_mean.shape # = (4096,)
    faces_repr = faces - faces_mean
    print "faces_repr.shape =", faces_repr.shape # = (400, 4096)

# (I have also tried to not center the face images, but the implementation of
# PCA used here seems to do that in any case, so that the results are not
# ignificantly different but the interpretation unclear.)

#------------------------------------------------------------------------------

# EXERCISE 2: When doing PCA, one normally centers the data first.  However, in
# some cases one might not want to do that and want the eigenvectors of the
# second moment matrix rather than of the covariance matrix.  The algorithm
# used here, however, does not seem to allow that, because it seems to
# implicitly or explicitly center the data itself (not documented).  What can
# you do to get the eigenvectors of the second moment matrix nevertheless?
# Implement your idea.

# SOLUTION:
# One can simply double the data points by adding their negative, i.e. append
# -faces to faces.  The data then has zero mean and the covariance matrix is
# identical to the original second moment matrix.
if not CENTER_FACES:
    # This dummy faces_mean is needed for the code further down.
    faces_mean = np.zeros(4096)
    print "faces_mean.shape =", faces_mean.shape # = (4096,)
    faces_repr = np.concatenate((faces, -faces), axis=0)
    print "faces_repr.shape =", faces_repr.shape # = (800, 4096)

#------------------------------------------------------------------------------

# EXERCISE 3: 
# (a): Visualize how faces look on the way between the mean face and the
# actual face.  
# (b): What do you think, which version looks prettiest?
# (c): Can you create caricatures by exaggerating the distance of a face from
# the mean face?

# SOLUTION:

# Ad (a):
animation = [faces_mean + dist/10.*faces_repr[id_idx] 
             for id_idx in range(0,50,10) for dist in range(11)]
             # Replace /10. by /5. to get beyond the original face.
show_gallery("From the mean face to individual faces for five different persons",
             animation, num_col=11, num_row=5)

# Ad (b): A while ago (guess when), there was a scientist who wanted to figure
# out how the typical murderer or thief or sexual criminal looks by averaging
# over many faces of each of these categories.  Surprisingly, the more he
# averaged the prettier and innocent the faces looked.  Do you see a similar
# effect here as you go from the original face (at the very right) towards the
# mean face (at the very left)?

# Ad (c): Creating a caricature does not work so well on this dataset (replace
# dist/10. above by dist/5. to get beyond the original face and see for
# yourself), but the method has indeed been used, see Brunelli, R. & Poggio,
# T. (1993) Caricatural effects in automated face perception.  Biological
# Cybernetics, 69(3), 235-241.
animation = [faces_mean + dist/5.*faces_repr[id_idx] 
             for id_idx in range(0,50,10) for dist in range(11)]
             # Replace /10. by /5. to get beyond the original face.
show_gallery("From the mean face to individual faces and beyond to caricatures for five different persons",
             animation, num_col=11, num_row=5)

#------------------------------------------------------------------------------

# EXERCISE 4: 
# (a): How many non-zero eigenvalues do you expect if you perform PCA on the
# centered face images?  Why?
# (b): What does that mean for the eigenvectors?

# SOLUTION:

# Ad (a): There are only 399 non-zero eigenvalues here, because there are only
# 400 face images, which can at most span a 400-dimensional space, which
# reduces to 399 if you center the images.

# Ad (b): The eigenvectors of the non-zero eigenvalues reflect the variances of
# the data in different directions and are in this sense meaningful.  The
# eigenvectors for zero eigenvalues would be random, if the algorithm would
# calculate them, apart from being orthonormal to earlier ones and among each
# other.

#------------------------------------------------------------------------------

# EXERCISE 5: 
# (a): Perform PCA on the centered face images.  
# (b): How many non-zero eigenvalues do you get?

# The eigenvectors of face images are often called eigenfaces.

# SOLUTION:

# Ad (a):
estimator = decomposition.PCA(n_components=400, svd_solver='auto', 
                              whiten=False)
estimator.fit(faces_repr)
eigenfaces = estimator.components_
eigenvalues = estimator.explained_variance_
print "eigenfaces.shape =", eigenfaces.shape # = (400, 4096)
print "eigenvalues.shape =", eigenvalues.shape # = (400,)

# Ad (b):
print eigenvalues
# The estimator does not seem to determine more eigenvectors than there are data
# points, which is quite reasonable.  The last eigenvalue though is close to
# zero, due to the centering.  Thus there are indeed 399 significantly non-zero
# eigenvalues.

#------------------------------------------------------------------------------

# EXERCISE 6: The PCA algorithm used here does not seem to determine more
# eigenvectors than there are data points.  Develop ideas (no need to implement
# that) how one could get the missing eigenvectors to get a full orthonormal
# basis?

# SOLUTION:
# To get a full orthonormal basis, one can complement the data with additional
# vectors with low noise to trick the PCA algorithm to generate a full set of
# eigenvectors.  The noise has to be small though not to spoil the meaningful
# eigenvectors. - Maybe a better possibility is to create some random vectors
# and then apply Gram-Schmidt orthonormalization on the eigenvectors plus the
# random vectors.  Since the eigenvectors are orthonormal already, Gram-Schmidt
# should not change them and only the random vectors are orthonormalized.  It is
# important though that Gram-Schmidt starts with the eigenvectors and not the
# random vectors.

#------------------------------------------------------------------------------

# EXERCISE 7: 
# (a) Plot the eigenvalue spectrum.
# (b) How many eigenfaces, do you think, are useful?
# (c) Would the eigenvalue spectrum look much different if you had many more
# face images?

# SOLUTION:

# Ad (a):
fig = plt.figure(figsize=(15, 10))
fig.suptitle("Eigenvalue spectrum", size=15)
ax = fig.add_subplot(111)
ax.scatter(range(400),eigenvalues)
ax.set(xlim=[-3,403],
       ylim=[0,1.1*np.max(eigenvalues)],
       xlabel='eigenface',
       ylabel='explained variance')
plt.show(block=SHOW_BLOCKS)
# It is also interesting to plot it on a log scale.

# Ad (b): Something like 50 to 100, I would say.  After that the eigenvalues
# are virtually zero.

# Ad (c): No, the number of face images should not change the spectrum in any
# systematic way, at least not the early values.  For late values it is quite
# clear that there are no more non-zero eigenvalues than there are face images.

#------------------------------------------------------------------------------

# EXERCISE 8: Plot the first 50 eigenfaces.

# SOLUTION:
# Since the sign of the eigenfaces is arbitrary, I plot the sign flipped
# version as well.
show_gallery("First fifty eigenfaces (sign not flipped)",  eigenfaces[:50], num_col=10, num_row=5)
show_gallery("First fifty eigenfaces (sign flipped)",   -eigenfaces[:50], num_col=10, num_row=5)

#------------------------------------------------------------------------------

# EXERCISE 9: Show the effect of an eigenface on the mean face, i.e. visualize
# the mean face plus different amounts of an eigenface.  Do this for different
# eigenfaces.

# SOLUTION:
animation = [faces_mean + std*np.sqrt(eigenvalues[rank])*eigenfaces[rank] 
             for rank in range(5) for std in range(-4,+5)]
show_gallery("Effect of eigenfaces on the mean faces (+-4 std)", 
             animation[:54], num_col=9, num_row=5)

#------------------------------------------------------------------------------

# EXERCISE 10: Compare the eigenfaces with and without centering the face
# images.  What do you observe?  Explain the differences and similarities.

# SOLUTION:
# The eigenfaces are remarkably similar.  Some are sign-flipped or swapped in
# rank, which does not mean much if the eigenvalues are similar.  Apart from
# that the similarity is remarkable.  A bit puzzling is the first eigenface.
# Not surprisingly the first eigenvector of the non-centered data looks exactly
# like the mean face.  That in turn is globally similar to the first eigenface
# of the centered data, which indicates that the main variation happens in the
# direction of the mean, which can be verified in the projection plot
# (animate_proj) of the non-centered data (please notice here the different
# scales).  On the other hand the first eigenfaces also differ in important
# details, most prominently the eyes, which have high contrast for the
# non-centered data and low contrast for the centered data.  The reason for
# that probably is, that eyes have a similar high contrast in all faces, so
# that there is high contrast in the mean but little variance across faces.  So
# far so good.  What is a bit surprising is that the other eigenfaces are so
# similar even though the first eigenface differs so much around the eyes.
# After all, the eigenfaces are orthonormal, and one would expect that a
# difference in one must be compensated by a difference in others.  Now, there
# are two possibilities.  The differences around the eyes might actually show
# up in other eigenfaces as well, we just have to look more closely, and/or the
# eyes are so similar between faces, that there is so little variance that this
# would only show up in later eigenfaces, which we do not consider so far.
# Calculating the angle and differences between the eigenfaces of centered and
# non-centered data could provide more insight here.

#------------------------------------------------------------------------------

# EXERCISE 11: Project face images into the space of eigenvectors.

# SOLUTION:
faces_proj = np.einsum('ij,kj->ik', faces_repr, eigenfaces)
print "faces_proj.shape =", faces_proj.shape # = (400, 4096)

#------------------------------------------------------------------------------

# EXERCISE 12: Reconstruct the faces from the projections.

# SOLUTION:
faces_rec = faces_mean + np.einsum('ji,ik->jk', faces_proj, eigenfaces)
print "faces_rec.shape =", faces_rec.shape # = (400, 4096)

#------------------------------------------------------------------------------

# EXERCISE 13: Plot all 10 views of the first five reconstructed faces in the
# data base.

# SOLUTION:
show_gallery("All views of first five reconstructed faces", faces_rec[:50],
             num_col=10, num_row=5)

#------------------------------------------------------------------------------

# EXERCISE 14: Show how the reconstruction depends on the number of eigenfaces used.

# SOLUTION:
animation = [faces_mean + np.einsum('i,ik->k', faces_proj[id_idx,:num_ef], eigenfaces[:num_ef]) 
             for id_idx in range(0,50,10) for num_ef in range(0,100,10)]
show_gallery("First five faces reconstructed with 0, 10, 20, ..., 90 eigenfaces",
             animation, num_col=10, num_row=5)

#------------------------------------------------------------------------------

# EXERCISE 15: Animate the face images of individual ids (=persons) projected
# onto the first two eigenfaces to see how they cluster per id.

# SOLUTION:
animate_proj(faces_proj, 0, 1)

#------------------------------------------------------------------------------

# EXERCISE 16: Imagine you have only one view per person in the gallery (but
# you were allowed to train PCA on all images for lack of more images).  Use
# the other views as probe images to perform face recognition.  A simple face
# recognition system based on nearest neighbor classification would search for
# the gallery image with the smallest Eucleadian distance to the probe image.
# Implement such a face recognition system on the raw images and the projected
# images.  Evaluate the performance and see how it depends on the number of
# eigenfaces used for the projection.  Maybe also check, how performance
# increases if you use more than one view per person in the gallery.

# SOLUTION:

# Define a subset of the 10 views [0, 1, ..., 9] as gallery images and a
# disjoint subset as probe images.
# This is for five views per person in the gallery.
gallery_view_set = [0,1,2,3,4]
probe_view_set = [5,6,7,8,9]
# This is for just one view per person in the gallery.
gallery_view_set = [0]
probe_view_set = [1,2,3,4,5,6,7,8,9]

def recognition_rate_nearest_neighbor_classification(faces_repr):
    """
    Routine to evaluate the performance of the nearest neighbor classifier on the faces.
    """

    # Go through all face images of the probe_set and try to recognize the
    # correct person in the gallery_set.  Determine the correct recognition
    # rate.
    num_trials = 0
    num_correct = 0
    for probe_id_idx in range(40):
        for probe_view_idx in probe_view_set:
            probe_idx = probe_id_idx*10 + probe_view_idx
            best_distance = None
            for gallery_id_idx in range(40):
                for gallery_view_idx in gallery_view_set:
                    gallery_idx = gallery_id_idx*10 + gallery_view_idx
                    distance = np.sqrt(np.sum((faces_repr[probe_idx]-faces_repr[gallery_idx])**2))
                    if distance < best_distance or best_distance == None:
                        best_distance = distance
                        best_gallery_id_idx = gallery_id_idx
            num_trials += 1
            if best_gallery_id_idx == probe_id_idx:
                num_correct += 1

    return float(num_correct)/num_trials

# Evaluate the nearest neighbor classifier on the raw images.
print "Recognition rate of nearest neighbor classifier on the raw images."
print recognition_rate_nearest_neighbor_classification(faces)

# Evaluate the nearest neighbor classifier on the projected images.
print "Recognition rate of nearest neighbor classifier on the images projected onto first 2, 4, 6, 8, ..., 398 PCs."
for pca_components in range(2,400,2):
    print pca_components, recognition_rate_nearest_neighbor_classification(faces_proj[:,:pca_components])

# The nearest neighbor classifier on the projected images works as well as the
# one on raw images with surprizingly few dimensions.  It even performs
# slightly better.

# PCA has actually been used for serious face recognition, see Turk, M. &
# Pentland, A. P. (1991) Eigenfaces for Recognition.  Journal of Cognitive
# Neuroscience, 3(1), 71-86.

#==============================================================================
#==============================================================================

# BSD 2-Clause License

# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

#==============================================================================

